"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Field = void 0;
const utils_1 = require("./utils");
const constants_1 = require("./constants");
class Field {
    constructor(parent, form, options) {
        this.options = {
            id: '',
            type: constants_1.FIELD_TYPE_TEXT,
            required: false,
            validation: (value, data, required) => {
                if (required && !value)
                    return constants_1.DEFAULT_REQUIRED_VALIDATION_MESSAGE;
                return true;
            },
            default: '',
            debounce: 200,
            className: constants_1.INPUT_CLASS_DEFAULT,
        };
        this.inputElement = null;
        this.containerElement = null;
        this.labelElement = null;
        this.validationElement = null;
        this.hasScheme = false;
        this._isValid = null;
        this._isDisabled = false;
        this._isRequired = false;
        this._isMounted = false;
        this._isVisible = true;
        this._vMessage = null;
        this._value = null;
        this._parsedConditions = null;
        this._parsedValidationConditions = null;
        this._parsedRequiredConditions = null;
        this._parsedDisabledConditions = null;
        this.initializeOptions(options);
        this._id = this.options.id;
        this._form = form;
        this._parent = parent;
        this._type = options.type;
        this._saveKey = (0, utils_1.generateFieldSaveKey)(this._form.getId(), this._id);
    }
    initializeOptions(options) {
        this.options = Object.assign({}, this.options, (0, utils_1.getOverwritenDefaults)(this.options.type, options));
    }
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            this.parseStringConditions();
            this.load();
            this.update();
            this.bindChange();
        });
    }
    parseStringConditions() {
        if (typeof this.options.conditions === 'string') {
            this._parsedConditions = (0, utils_1.parseConditionString)(this.options.conditions);
        }
        if (typeof this.options.disabled === 'string') {
            this._parsedDisabledConditions = (0, utils_1.parseConditionString)(this.options.disabled);
        }
        if (typeof this.options.required === 'string') {
            this._parsedRequiredConditions = (0, utils_1.parseConditionString)(this.options.required);
        }
        if (typeof this.options.validation === 'string') {
            this._parsedValidationConditions = (0, utils_1.parseConditionString)(this.options.validation);
        }
    }
    setValue(value, save = true) {
        this._value = value;
        this.syncValue();
        this._form.setData(this._id, value);
        if (save)
            this.save();
    }
    syncValue() {
        if (this.inputElement &&
            (this.inputElement instanceof HTMLInputElement || this.inputElement instanceof HTMLTextAreaElement) &&
            this.inputElement.value !== this._value) {
            this.inputElement.value = this._value;
        }
    }
    getValue() {
        return this._value;
    }
    getId() {
        return this._id;
    }
    getSaveKey() {
        return this._saveKey;
    }
    getType() {
        return this._type;
    }
    getVisibility() {
        return this._isVisible;
    }
    getValidity() {
        return this._isValid;
    }
    isDisabled() {
        return this._isDisabled;
    }
    isRequired() {
        return this._isRequired;
    }
    getValidationMessage() {
        return this._vMessage;
    }
    getForm() {
        return this._form;
    }
    createContainerElement() {
        this.containerElement = document.createElement('div');
        this.containerElement.className = constants_1.FIELD_CLASS_DEFAULT + ' ' + this._type;
        this.containerElement.setAttribute(constants_1.ID_ATTRIBUTE, this._id + constants_1.CONTAINER_DEFINITION);
    }
    createInputElement() {
        this.inputElement = document.createElement(constants_1.INPUT_ELEMENT);
        this.inputElement.setAttribute(constants_1.ID_ATTRIBUTE, this._id);
        this.inputElement.setAttribute(constants_1.NAME_ATTRIBUTE, this.options.name || this._id);
        this.inputElement.setAttribute(constants_1.TYPE_ATTRIBUTE, this._type);
        if (this.options.placeholder)
            this.inputElement.setAttribute(constants_1.PLACEHOLDER_ATTRIBUTE, this.options.placeholder);
        this.inputElement.className = this.options.className;
    }
    createLabelElement() {
        this.labelElement = document.createElement(constants_1.LABEL_ELEMENT);
        const label = document.createElement(constants_1.PARAGRAPH_ELEMENT);
        if (this.options.label)
            label.innerText = this.options.label;
        label.setAttribute(constants_1.ID_ATTRIBUTE, this._id + constants_1.LABEL_DEFINITION);
        this.labelElement.setAttribute(constants_1.FOR_ATTRIBUTE, this._id);
        label.className = constants_1.LABEL_CLASS_DEFAULT;
        if (this.options.label)
            (0, utils_1.mountElement)(label, this.labelElement);
    }
    createValidationElement() {
        this.validationElement = document.createElement(constants_1.PARAGRAPH_ELEMENT);
        this.validationElement.setAttribute(constants_1.ID_ATTRIBUTE, this._id + constants_1.VALIDATION_DEFINITION);
        this.validationElement.className = constants_1.VALIDATION_CLASS_DEFAULT;
        this.validationElement.style.display = 'none';
    }
    onGui() {
        this.createContainerElement();
        this.createInputElement();
        this.createLabelElement();
        this.createValidationElement();
        if (this.inputElement && this.labelElement)
            (0, utils_1.mountElement)(this.inputElement, this.labelElement);
        if (this.labelElement && this.validationElement)
            (0, utils_1.mountElement)(this.validationElement, this.labelElement);
        if (this.containerElement && this.labelElement)
            (0, utils_1.mountElement)(this.labelElement, this.containerElement);
    }
    mount() {
        if (this.containerElement)
            (0, utils_1.mountElement)(this.containerElement, this._parent);
        this._isMounted = true;
    }
    unmount() {
        if (this.containerElement)
            (0, utils_1.unmountElement)(this.containerElement);
        this._isMounted = false;
    }
    destroy() {
        if (this._parent)
            this._parent.remove();
    }
    handleValidatedField() {
        var _a, _b, _c, _d, _e, _f, _g;
        if (!this._isValid) {
            (_a = this.labelElement) === null || _a === void 0 ? void 0 : _a.classList.add(constants_1.FORM_ERROR_CLASS_DEFAULT);
            if (this.validationElement && this._vMessage) {
                this.validationElement.innerText = this._vMessage;
                this.validationElement.style.display = 'block';
            }
            (_b = this.inputElement) === null || _b === void 0 ? void 0 : _b.setAttribute(constants_1.ARIA_INVALID_ATTRIBUTE, 'true');
            (_c = this.inputElement) === null || _c === void 0 ? void 0 : _c.setAttribute(constants_1.ARIA_DESCRIBEDBY_ATTRIBUTE, ((_d = this.validationElement) === null || _d === void 0 ? void 0 : _d.getAttribute('id')) || '');
        }
        else {
            if (this.validationElement) {
                this.validationElement.style.display = 'none';
            }
            (_e = this.labelElement) === null || _e === void 0 ? void 0 : _e.classList.remove(constants_1.FORM_ERROR_CLASS_DEFAULT);
            (_f = this.inputElement) === null || _f === void 0 ? void 0 : _f.removeAttribute(constants_1.ARIA_INVALID_ATTRIBUTE);
            (_g = this.inputElement) === null || _g === void 0 ? void 0 : _g.removeAttribute(constants_1.ARIA_DESCRIBEDBY_ATTRIBUTE);
        }
        this._form.updateError(this._id, this._isValid);
    }
    handleVisibility() {
        if (this._isVisible && !this._isMounted) {
            this.mount();
            this._form.setData(this._id, this._value);
        }
        if (!this._isVisible && this._isMounted) {
            this._form.removeData(this._id);
            this.unmount();
        }
    }
    handleDisabled() {
        var _a, _b, _c, _d;
        if (this._isDisabled) {
            (_a = this.containerElement) === null || _a === void 0 ? void 0 : _a.classList.add(constants_1.FIELD_DISABLED_CLASS_DEFAULT);
            (_b = this.inputElement) === null || _b === void 0 ? void 0 : _b.setAttribute(constants_1.DISABLED_ATTRIBUTE, 'true');
        }
        else {
            (_c = this.inputElement) === null || _c === void 0 ? void 0 : _c.removeAttribute(constants_1.DISABLED_ATTRIBUTE);
            (_d = this.containerElement) === null || _d === void 0 ? void 0 : _d.classList.remove(constants_1.FIELD_DISABLED_CLASS_DEFAULT);
        }
    }
    handleRequired() {
        var _a, _b;
        if (this._isRequired) {
            (_a = this.containerElement) === null || _a === void 0 ? void 0 : _a.classList.add(constants_1.FIELD_REQUIRED_CLASS_DEFAULT);
        }
        else {
            (_b = this.containerElement) === null || _b === void 0 ? void 0 : _b.classList.remove(constants_1.FIELD_REQUIRED_CLASS_DEFAULT);
        }
    }
    bindChange() {
        if (this.inputElement)
            this.inputElement.addEventListener(constants_1.INPUT_ATTRIBUTE, (0, utils_1.debounce)(this.change, this.options.debounce, this));
    }
    save() {
        if (this._form.savesProgress() && this._form.hasValidLicense() && this._value !== undefined) {
            localStorage.setItem(this._saveKey, JSON.stringify(this._value));
        }
    }
    load() {
        var _a;
        if (this._form.savesProgress() && this._form.hasValidLicense()) {
            const value = localStorage.getItem(this._saveKey);
            if (value !== null) {
                this.setValue((0, utils_1.isJson)(value) ? JSON.parse(value) : value, false);
                return;
            }
        }
        this.setValue((_a = this.options.default) !== null && _a !== void 0 ? _a : null, false);
    }
    reset() {
        return __awaiter(this, void 0, void 0, function* () {
            localStorage.removeItem(this._saveKey);
            if (this.options.default !== undefined)
                this.setValue(this.options.default, false);
            else
                this.setValue(null, false);
            this.update();
        });
    }
    update() {
        return __awaiter(this, void 0, void 0, function* () {
            this.updateVisibilityBasedOnConditions();
            this.updateDisabledStatus();
            this.updateRequiredStatus();
            this.handleVisibility();
            this.handleDisabled();
            this.handleRequired();
        });
    }
    updateVisibilityBasedOnConditions() {
        if (this.options.conditions) {
            if (this._parsedConditions) {
                this._isVisible = (0, utils_1.evaluateParsedConditions)(this._parsedConditions, this._form.getData(), this._value, this._isRequired);
            }
            else if (typeof this.options.conditions === 'function') {
                this._isVisible = this.options.conditions(this._value, this._form.getData());
            }
        }
    }
    updateDisabledStatus() {
        if (this.options.disabled) {
            if (typeof this.options.disabled === 'string' && this._parsedDisabledConditions) {
                this._isDisabled = (0, utils_1.evaluateParsedConditions)(this._parsedDisabledConditions, this._form.getData(), this._value, this._isRequired);
            }
            else {
                if (typeof this.options.disabled === 'function') {
                    this._isDisabled = this.options.disabled(this._value, this._form.getData());
                }
                else if (typeof this.options.disabled === 'boolean') {
                    this._isDisabled = this.options.disabled;
                }
            }
        }
    }
    updateRequiredStatus() {
        if (this.options.required) {
            if (typeof this.options.required === 'string' && this._parsedRequiredConditions) {
                this._isRequired = (0, utils_1.evaluateParsedConditions)(this._parsedRequiredConditions, this._form.getData(), this._value);
            }
            else {
                if (typeof this.options.required === 'function') {
                    this._isRequired = this.options.required(this._value, this._form.getData());
                }
                else if (typeof this.options.required === 'boolean') {
                    this._isRequired = this.options.required;
                }
            }
        }
    }
    validate() {
        if (!this._isVisible)
            return true;
        if (this.options.validation) {
            if (typeof this.options.validation === 'string' && this._parsedValidationConditions) {
                const validity = (0, utils_1.evaluateParsedConditions)(this._parsedValidationConditions, this._form.getData(), this._value, this._isRequired, true);
                this.setValidationValues(validity);
            }
            else if (typeof this.options.validation === 'function') {
                const validity = this.options.validation(this._value, this._form.getData(), this._isRequired);
                this.setValidationValues(validity);
            }
            else
                this._isValid = true;
        }
        else
            this._isValid = true;
        return this._isValid;
    }
    setValidationValues(validity) {
        this._isValid = validity === true;
        this._vMessage = validity === true ? '' : validity;
    }
    change(event) {
        this.setValue(event.target.value);
        this.validate();
        this.handleValidatedField();
        if (this.options.change)
            this.options.change(this._value);
    }
}
exports.Field = Field;
