import { debounce, evaluateParsedConditions, generateFieldSaveKey, getOverwritenDefaults, isJson, mountElement, parseConditionString, unmountElement, } from './utils';
import { ARIA_DESCRIBEDBY_ATTRIBUTE, ARIA_INVALID_ATTRIBUTE, CONTAINER_DEFINITION, DEFAULT_REQUIRED_VALIDATION_MESSAGE, DISABLED_ATTRIBUTE, FIELD_CLASS_DEFAULT, FIELD_DISABLED_CLASS_DEFAULT, FIELD_REQUIRED_CLASS_DEFAULT, FIELD_TYPE_TEXT, FORM_ERROR_CLASS_DEFAULT, FOR_ATTRIBUTE, ID_ATTRIBUTE, INPUT_ATTRIBUTE, INPUT_CLASS_DEFAULT, INPUT_ELEMENT, LABEL_CLASS_DEFAULT, LABEL_DEFINITION, LABEL_ELEMENT, NAME_ATTRIBUTE, PARAGRAPH_ELEMENT, PLACEHOLDER_ATTRIBUTE, TYPE_ATTRIBUTE, VALIDATION_CLASS_DEFAULT, VALIDATION_DEFINITION, } from './constants';
export class Field {
    options = {
        id: '',
        type: FIELD_TYPE_TEXT,
        required: false,
        validation: (value, data, required) => {
            if (required && !value)
                return DEFAULT_REQUIRED_VALIDATION_MESSAGE;
            return true;
        },
        default: '',
        debounce: 200,
        className: INPUT_CLASS_DEFAULT,
    };
    inputElement = null;
    containerElement = null;
    labelElement = null;
    validationElement = null;
    hasScheme = false;
    _id;
    _parent;
    _form;
    _isValid = null;
    _isDisabled = false;
    _isRequired = false;
    _isMounted = false;
    _isVisible = true;
    _saveKey;
    _vMessage = null;
    _value = null;
    _type;
    _parsedConditions = null;
    _parsedValidationConditions = null;
    _parsedRequiredConditions = null;
    _parsedDisabledConditions = null;
    constructor(parent, form, options) {
        this.initializeOptions(options);
        this._id = this.options.id;
        this._form = form;
        this._parent = parent;
        this._type = options.type;
        this._saveKey = generateFieldSaveKey(this._form.getId(), this._id);
    }
    initializeOptions(options) {
        this.options = Object.assign({}, this.options, getOverwritenDefaults(this.options.type, options));
    }
    async initialize() {
        this.parseStringConditions();
        this.load();
        this.update();
        this.bindChange();
    }
    parseStringConditions() {
        if (typeof this.options.conditions === 'string') {
            this._parsedConditions = parseConditionString(this.options.conditions);
        }
        if (typeof this.options.disabled === 'string') {
            this._parsedDisabledConditions = parseConditionString(this.options.disabled);
        }
        if (typeof this.options.required === 'string') {
            this._parsedRequiredConditions = parseConditionString(this.options.required);
        }
        if (typeof this.options.validation === 'string') {
            this._parsedValidationConditions = parseConditionString(this.options.validation);
        }
    }
    setValue(value, save = true) {
        this._value = value;
        this.syncValue();
        this._form.setData(this._id, value);
        if (save)
            this.save();
    }
    syncValue() {
        if (this.inputElement &&
            (this.inputElement instanceof HTMLInputElement || this.inputElement instanceof HTMLTextAreaElement) &&
            this.inputElement.value !== this._value) {
            this.inputElement.value = this._value;
        }
    }
    getValue() {
        return this._value;
    }
    getId() {
        return this._id;
    }
    getSaveKey() {
        return this._saveKey;
    }
    getType() {
        return this._type;
    }
    getVisibility() {
        return this._isVisible;
    }
    getValidity() {
        return this._isValid;
    }
    isDisabled() {
        return this._isDisabled;
    }
    isRequired() {
        return this._isRequired;
    }
    getValidationMessage() {
        return this._vMessage;
    }
    getForm() {
        return this._form;
    }
    createContainerElement() {
        this.containerElement = document.createElement('div');
        this.containerElement.className = FIELD_CLASS_DEFAULT + ' ' + this._type;
        this.containerElement.setAttribute(ID_ATTRIBUTE, this._id + CONTAINER_DEFINITION);
    }
    createInputElement() {
        this.inputElement = document.createElement(INPUT_ELEMENT);
        this.inputElement.setAttribute(ID_ATTRIBUTE, this._id);
        this.inputElement.setAttribute(NAME_ATTRIBUTE, this.options.name || this._id);
        this.inputElement.setAttribute(TYPE_ATTRIBUTE, this._type);
        if (this.options.placeholder)
            this.inputElement.setAttribute(PLACEHOLDER_ATTRIBUTE, this.options.placeholder);
        this.inputElement.className = this.options.className;
    }
    createLabelElement() {
        this.labelElement = document.createElement(LABEL_ELEMENT);
        const label = document.createElement(PARAGRAPH_ELEMENT);
        if (this.options.label)
            label.innerText = this.options.label;
        label.setAttribute(ID_ATTRIBUTE, this._id + LABEL_DEFINITION);
        this.labelElement.setAttribute(FOR_ATTRIBUTE, this._id);
        label.className = LABEL_CLASS_DEFAULT;
        if (this.options.label)
            mountElement(label, this.labelElement);
    }
    createValidationElement() {
        this.validationElement = document.createElement(PARAGRAPH_ELEMENT);
        this.validationElement.setAttribute(ID_ATTRIBUTE, this._id + VALIDATION_DEFINITION);
        this.validationElement.className = VALIDATION_CLASS_DEFAULT;
        this.validationElement.style.display = 'none';
    }
    onGui() {
        this.createContainerElement();
        this.createInputElement();
        this.createLabelElement();
        this.createValidationElement();
        if (this.inputElement && this.labelElement)
            mountElement(this.inputElement, this.labelElement);
        if (this.labelElement && this.validationElement)
            mountElement(this.validationElement, this.labelElement);
        if (this.containerElement && this.labelElement)
            mountElement(this.labelElement, this.containerElement);
    }
    mount() {
        if (this.containerElement)
            mountElement(this.containerElement, this._parent);
        this._isMounted = true;
    }
    unmount() {
        if (this.containerElement)
            unmountElement(this.containerElement);
        this._isMounted = false;
    }
    destroy() {
        if (this._parent)
            this._parent.remove();
    }
    handleValidatedField() {
        if (!this._isValid) {
            this.labelElement?.classList.add(FORM_ERROR_CLASS_DEFAULT);
            if (this.validationElement && this._vMessage) {
                this.validationElement.innerText = this._vMessage;
                this.validationElement.style.display = 'block';
            }
            this.inputElement?.setAttribute(ARIA_INVALID_ATTRIBUTE, 'true');
            this.inputElement?.setAttribute(ARIA_DESCRIBEDBY_ATTRIBUTE, this.validationElement?.getAttribute('id') || '');
        }
        else {
            if (this.validationElement) {
                this.validationElement.style.display = 'none';
            }
            this.labelElement?.classList.remove(FORM_ERROR_CLASS_DEFAULT);
            this.inputElement?.removeAttribute(ARIA_INVALID_ATTRIBUTE);
            this.inputElement?.removeAttribute(ARIA_DESCRIBEDBY_ATTRIBUTE);
        }
        this._form.updateError(this._id, this._isValid);
    }
    handleVisibility() {
        if (this._isVisible && !this._isMounted) {
            this.mount();
            this._form.setData(this._id, this._value);
        }
        if (!this._isVisible && this._isMounted) {
            this._form.removeData(this._id);
            this.unmount();
        }
    }
    handleDisabled() {
        if (this._isDisabled) {
            this.containerElement?.classList.add(FIELD_DISABLED_CLASS_DEFAULT);
            this.inputElement?.setAttribute(DISABLED_ATTRIBUTE, 'true');
        }
        else {
            this.inputElement?.removeAttribute(DISABLED_ATTRIBUTE);
            this.containerElement?.classList.remove(FIELD_DISABLED_CLASS_DEFAULT);
        }
    }
    handleRequired() {
        if (this._isRequired) {
            this.containerElement?.classList.add(FIELD_REQUIRED_CLASS_DEFAULT);
        }
        else {
            this.containerElement?.classList.remove(FIELD_REQUIRED_CLASS_DEFAULT);
        }
    }
    bindChange() {
        if (this.inputElement)
            this.inputElement.addEventListener(INPUT_ATTRIBUTE, debounce(this.change, this.options.debounce, this));
    }
    save() {
        if (this._form.savesProgress() && this._form.hasValidLicense() && this._value !== undefined) {
            localStorage.setItem(this._saveKey, JSON.stringify(this._value));
        }
    }
    load() {
        if (this._form.savesProgress() && this._form.hasValidLicense()) {
            const value = localStorage.getItem(this._saveKey);
            if (value !== null) {
                this.setValue(isJson(value) ? JSON.parse(value) : value, false);
                return;
            }
        }
        this.setValue(this.options.default ?? null, false);
    }
    async reset() {
        localStorage.removeItem(this._saveKey);
        if (this.options.default !== undefined)
            this.setValue(this.options.default, false);
        else
            this.setValue(null, false);
        this.update();
    }
    async update() {
        this.updateVisibilityBasedOnConditions();
        this.updateDisabledStatus();
        this.updateRequiredStatus();
        this.handleVisibility();
        this.handleDisabled();
        this.handleRequired();
    }
    updateVisibilityBasedOnConditions() {
        if (this.options.conditions) {
            if (this._parsedConditions) {
                this._isVisible = evaluateParsedConditions(this._parsedConditions, this._form.getData(), this._value, this._isRequired);
            }
            else if (typeof this.options.conditions === 'function') {
                this._isVisible = this.options.conditions(this._value, this._form.getData());
            }
        }
    }
    updateDisabledStatus() {
        if (this.options.disabled) {
            if (typeof this.options.disabled === 'string' && this._parsedDisabledConditions) {
                this._isDisabled = evaluateParsedConditions(this._parsedDisabledConditions, this._form.getData(), this._value, this._isRequired);
            }
            else {
                if (typeof this.options.disabled === 'function') {
                    this._isDisabled = this.options.disabled(this._value, this._form.getData());
                }
                else if (typeof this.options.disabled === 'boolean') {
                    this._isDisabled = this.options.disabled;
                }
            }
        }
    }
    updateRequiredStatus() {
        if (this.options.required) {
            if (typeof this.options.required === 'string' && this._parsedRequiredConditions) {
                this._isRequired = evaluateParsedConditions(this._parsedRequiredConditions, this._form.getData(), this._value);
            }
            else {
                if (typeof this.options.required === 'function') {
                    this._isRequired = this.options.required(this._value, this._form.getData());
                }
                else if (typeof this.options.required === 'boolean') {
                    this._isRequired = this.options.required;
                }
            }
        }
    }
    validate() {
        if (!this._isVisible)
            return true;
        if (this.options.validation) {
            if (typeof this.options.validation === 'string' && this._parsedValidationConditions) {
                const validity = evaluateParsedConditions(this._parsedValidationConditions, this._form.getData(), this._value, this._isRequired, true);
                this.setValidationValues(validity);
            }
            else if (typeof this.options.validation === 'function') {
                const validity = this.options.validation(this._value, this._form.getData(), this._isRequired);
                this.setValidationValues(validity);
            }
            else
                this._isValid = true;
        }
        else
            this._isValid = true;
        return this._isValid;
    }
    setValidationValues(validity) {
        this._isValid = validity === true;
        this._vMessage = validity === true ? '' : validity;
    }
    change(event) {
        this.setValue(event.target.value);
        this.validate();
        this.handleValidatedField();
        if (this.options.change)
            this.options.change(this._value);
    }
}
