import { generateFieldSaveKey, getOverwritenDefaults, isJson, mountElement } from '../utils';
import { CONTAINER_DEFINITION, DIV_ELEMENT, FIELD_CLASS_DEFAULT, FIELD_TYPE_HIDDEN, ID_ATTRIBUTE, INPUT_ELEMENT, NAME_ATTRIBUTE, TYPE_ATTRIBUTE, } from '../constants';
export class HiddenField {
    options = {
        id: '',
        type: FIELD_TYPE_HIDDEN,
        default: null,
    };
    inputElement = null;
    containerElement = null;
    labelElement = null;
    validationElement = null;
    _id;
    _parent;
    _form;
    _saveKey;
    _value = null;
    _type;
    constructor(parent, form, options) {
        this.initializeOptions(options);
        this._id = this.options.id;
        this._form = form;
        this._parent = parent;
        this._type = options.type;
        this._saveKey = generateFieldSaveKey(this._form.getId(), this._id);
        this.onGui();
        this.initialize();
    }
    initializeOptions(options) {
        this.options = Object.assign({}, this.options, getOverwritenDefaults(this.options.type, options));
    }
    async initialize() {
        this.load();
        this.update();
    }
    setValue(value, save = true) {
        this._value = value;
        this.syncValue();
        this._form.setData(this._id, value);
        if (save)
            this.save();
    }
    getValue() {
        return this._value;
    }
    syncValue() {
        if (this.inputElement && this.inputElement instanceof HTMLInputElement && this.inputElement.value !== this._value) {
            this.inputElement.value = this._value;
        }
    }
    mountToParent() {
        if (this.containerElement)
            mountElement(this.containerElement, this._parent);
    }
    createContainerElement() {
        this.containerElement = document.createElement(DIV_ELEMENT);
        this.containerElement.className = FIELD_CLASS_DEFAULT + ' ' + this.getType();
        this.containerElement.setAttribute(ID_ATTRIBUTE, this.getId() + CONTAINER_DEFINITION);
    }
    createInputElement() {
        this.inputElement = document.createElement(INPUT_ELEMENT);
        this.inputElement.setAttribute(ID_ATTRIBUTE, this.getId());
        this.inputElement.setAttribute(NAME_ATTRIBUTE, this.options.name || this.getId());
        this.inputElement.setAttribute(TYPE_ATTRIBUTE, this.getType());
    }
    getId() {
        return this._id;
    }
    getSaveKey() {
        return this._saveKey;
    }
    getType() {
        return this._type;
    }
    getForm() {
        return this._form;
    }
    onGui() {
        this.createContainerElement();
        this.createInputElement();
        if (this.containerElement && this.inputElement)
            mountElement(this.inputElement, this.containerElement);
        this.mountToParent();
    }
    destroy() {
        if (this._parent)
            this._parent.remove();
    }
    save() {
        if (this._form.savesProgress() && this._form.hasValidLicense() && this._value !== undefined) {
            localStorage.setItem(this._saveKey, JSON.stringify(this._value));
        }
    }
    load() {
        if (this._form.savesProgress() && this._form.hasValidLicense()) {
            const value = localStorage.getItem(this._saveKey);
            if (value !== null) {
                this.setValue(isJson(value) ? JSON.parse(value) : value, false);
                return;
            }
        }
        this.setValue(this.options.default ?? null, false);
    }
    async reset() {
        localStorage.removeItem(this._saveKey);
        this.setValue(this.options.default ?? null, false);
        this.update();
    }
    async update() {
        return;
    }
}
