import { evaluateParsedConditions, generateFieldSaveKey, getOverwritenDefaults, isJson, mountElement, parseConditionString, unmountElement, } from '../utils';
import { CONTAINER_DEFINITION, DIV_ELEMENT, FIELD_CLASS_DEFAULT, FIELD_TYPE_STATIC, ID_ATTRIBUTE, STATIC_CLASS_DEFAULT, } from '../constants';
export class StaticField {
    options = {
        id: '',
        type: FIELD_TYPE_STATIC,
        template: '',
        className: STATIC_CLASS_DEFAULT,
    };
    staticElement = null;
    containerElement = null;
    _id;
    _parent;
    _form;
    _isMounted = false;
    _isVisible = true;
    _saveKey;
    _value = null;
    _type;
    _parsedConditions = null;
    constructor(parent, form, options) {
        this.initializeOptions(options);
        this._id = this.options.id;
        this._form = form;
        this._parent = parent;
        this._type = options.type;
        this._saveKey = generateFieldSaveKey(this._form.getId(), this._id);
        this.onGui();
        this.initialize();
    }
    initializeOptions(options) {
        this.options = Object.assign({}, this.options, getOverwritenDefaults(this.options.type, options));
    }
    async initialize() {
        this.parseStringConditions();
        this.load();
        this.update();
    }
    parseStringConditions() {
        if (typeof this.options.conditions === 'string') {
            this._parsedConditions = parseConditionString(this.options.conditions);
        }
    }
    getValue() {
        return this._value;
    }
    getId() {
        return this._id;
    }
    getSaveKey() {
        return this._saveKey;
    }
    getType() {
        return this._type;
    }
    getVisibility() {
        return this._isVisible;
    }
    getForm() {
        return this._form;
    }
    setTemplate(template, save = true) {
        this._value = template;
        if (this.staticElement)
            this.staticElement.innerHTML = template;
        if (save)
            this.save();
    }
    createContainerElement() {
        this.containerElement = document.createElement(DIV_ELEMENT);
        this.containerElement.className = FIELD_CLASS_DEFAULT + ' ' + this._type;
        this.containerElement.setAttribute(ID_ATTRIBUTE, this._id + CONTAINER_DEFINITION);
    }
    createStaticElement() {
        this.staticElement = document.createElement(DIV_ELEMENT);
        this.staticElement.className = this.options.className;
        this.staticElement.setAttribute(ID_ATTRIBUTE, this._id);
    }
    onGui() {
        this.createContainerElement();
        this.createStaticElement();
        if (this.containerElement && this.staticElement)
            mountElement(this.staticElement, this.containerElement);
    }
    mount() {
        if (this.containerElement)
            mountElement(this.containerElement, this._parent);
        this._isMounted = true;
    }
    unmount() {
        if (this.containerElement)
            unmountElement(this.containerElement);
        this._isMounted = false;
    }
    destroy() {
        if (this._parent)
            this._parent.remove();
    }
    handleVisibility() {
        if (this._isVisible && !this._isMounted)
            this.mount();
        if (!this._isVisible && this._isMounted)
            this.unmount();
    }
    save() {
        if (this._form.savesProgress() && this._form.hasValidLicense() && this._value !== undefined) {
            localStorage.setItem(this._saveKey, JSON.stringify(this._value));
        }
    }
    load() {
        if (this._form.savesProgress() && this._form.hasValidLicense()) {
            const value = localStorage.getItem(this._saveKey);
            if (value !== null) {
                this.setTemplate(isJson(value) ? JSON.parse(value) : value, false);
                return;
            }
        }
        this.setTemplate(this.options.template, false);
    }
    async reset() {
        localStorage.removeItem(this._saveKey);
        this.setTemplate(this.options.template, false);
        this.update();
    }
    async update() {
        this.updateVisibilityBasedOnConditions();
        this.handleVisibility();
    }
    updateVisibilityBasedOnConditions() {
        if (this.options.conditions) {
            if (this._parsedConditions) {
                this._isVisible = evaluateParsedConditions(this._parsedConditions, this._form.getData());
            }
            else if (typeof this.options.conditions === 'function') {
                this._isVisible = this.options.conditions(this.getValue(), this._form.getData());
            }
        }
    }
}
