import { DIV_ELEMENT, FIELD_KEY_DEFINITION, FORMSJS_KEY_DEFINITION, INVALID_CONSOLE_TEXT, INVALID_LICENSE_TEXT, LICENSE_STATE, OS_LICENSE_KEYS, OUTDATED_CONSOLE_TEXT, OUTDATED_LICENSE_TEXT, RELEASE_DATE, STRING_CONDITION_DATE_IDENTIFIER, STRING_CONDITION_SPLIT, STRING_CONDITION_VALUE_SPLIT, UPGRADE_WINDOW, VALID_LICENSE_TEXT, buttons, constructorTypes, fields, groups, } from './constants';
let LICENSE_KEY = null;
let USES_LICENSED_FETURES = false;
const OVEWRITE_DEFAULT_OPTIONS = {};
export const debounce = (callback, wait, context) => {
    let timeoutId = null;
    return (...args) => {
        if (timeoutId)
            window.clearTimeout(timeoutId);
        timeoutId = window.setTimeout(() => {
            callback.apply(context, args);
        }, wait);
    };
};
export const mountElement = (element, parent) => {
    parent.append(element);
};
export const unmountElement = (element) => {
    element.parentNode?.removeChild(element);
};
export const objectToFormData = (obj, form, namespace) => {
    const fd = form || new FormData();
    let formKey;
    for (const property in obj) {
        if (Object.prototype.hasOwnProperty.call(obj, property)) {
            if (namespace) {
                formKey = namespace + '[' + property + ']';
            }
            else {
                formKey = property;
            }
            if (obj[property] == null) {
                fd.append(formKey, '');
            }
            else if (typeof obj[property] === 'object' && !(obj[property] instanceof File)) {
                objectToFormData(obj[property], fd, property);
            }
            else {
                fd.append(formKey, obj[property]);
            }
        }
    }
    return fd;
};
export const isField = (type) => {
    return fields.includes(type);
};
export const isGroup = (type) => {
    return groups.includes(type);
};
export const isButton = (type) => {
    return buttons.includes(type);
};
export const getFormElementType = (type) => {
    if (isField(type))
        return constructorTypes.field;
    if (isGroup(type))
        return constructorTypes.group;
    if (isButton(type))
        return constructorTypes.button;
    return null;
};
export const extractFieldsFromSchema = (schema) => {
    let fieldsList = [];
    schema.forEach((options) => {
        if (fields.includes(options.type))
            fieldsList.push(options.id);
        else if (options.schema)
            fieldsList = [...fieldsList, ...extractFieldsFromSchema(options.schema)];
    });
    return fieldsList;
};
export const createWrapper = (parent) => {
    const wrapper = document.createElement(DIV_ELEMENT);
    mountElement(wrapper, parent);
    return wrapper;
};
export const transformFieldName = (parentId, key, id) => {
    return parentId + '[' + key + ']' + id;
};
export const generateFieldSaveKey = (formId, id) => {
    return FORMSJS_KEY_DEFINITION + formId + FIELD_KEY_DEFINITION + id;
};
const isDateValid = (d) => d instanceof Date && !isNaN(d);
const calculateDateAfterAddingDays = (date, days) => {
    const result = new Date(date);
    result.setDate(result.getDate() + days);
    return result;
};
const parseKey = (key) => key.match(/^(\d+)@(\d+)$/);
const isKeyInLicenseKeys = (key) => OS_LICENSE_KEYS.includes(key);
const calculateMinPurchaseDate = (releaseDate) => calculateDateAfterAddingDays(releaseDate, -UPGRADE_WINDOW);
const determineLicenseState = (purchaseDate, minPurchaseDate) => minPurchaseDate < purchaseDate ? LICENSE_STATE.VALID : LICENSE_STATE.OUTDATED;
export const processLicenseKey = () => {
    const key = LICENSE_KEY;
    if (!key)
        return LICENSE_STATE.INVALID;
    if (isKeyInLicenseKeys(key))
        return LICENSE_STATE.VALID;
    const parts = parseKey(key);
    if (parts?.[1]?.length === 10) {
        const [, , timestamp] = parts;
        const purchaseDate = new Date(parseInt(timestamp, 10) * 1000);
        const releaseDate = new Date(RELEASE_DATE);
        if (isDateValid(releaseDate)) {
            const minPurchaseDate = calculateMinPurchaseDate(releaseDate);
            return determineLicenseState(purchaseDate, minPurchaseDate);
        }
    }
    return LICENSE_STATE.INVALID;
};
export const setLicenseKey = (key) => {
    LICENSE_KEY = key;
};
export const getLicenseText = (license) => {
    if (license === LICENSE_STATE.VALID)
        return VALID_LICENSE_TEXT;
    if (license === LICENSE_STATE.OUTDATED)
        return OUTDATED_LICENSE_TEXT;
    return INVALID_LICENSE_TEXT;
};
export const handleInvalidLicenseLog = (license) => {
    if (license === LICENSE_STATE.INVALID) {
        console.error(INVALID_CONSOLE_TEXT);
    }
    else {
        console.error(OUTDATED_CONSOLE_TEXT);
    }
};
export const usesLicensedFetures = () => {
    return USES_LICENSED_FETURES;
};
export const useLicensedFetures = () => {
    USES_LICENSED_FETURES = true;
};
export const parseConditionString = (conditionStr) => {
    return conditionStr
        .split(STRING_CONDITION_SPLIT)
        .filter((part) => part.includes(STRING_CONDITION_VALUE_SPLIT))
        .map((part) => {
        const [rawConditions, returnValue] = part.split(STRING_CONDITION_VALUE_SPLIT);
        const conditions = rawConditions
            .replace(/^\[|\]$/g, '')
            .split('&&')
            .map((andCond) => andCond.split('||').map((orCond) => {
            const match = orCond.match(/(.*?)(=|!=|>|<|>=|<=)(.*)/);
            if (!match)
                throw new Error(`Invalid condition format: ${orCond}`);
            const [, left, operator, rightPart] = match;
            let right;
            const isDate = rightPart.trim().startsWith(STRING_CONDITION_DATE_IDENTIFIER);
            if (isDate) {
                right = new Date(rightPart.trim().substring(5));
            }
            else {
                try {
                    right = JSON.parse(rightPart.trim());
                }
                catch {
                    right = rightPart.trim();
                }
            }
            return { left: left.trim(), operator: operator, right, isDate };
        }));
        return { conditions, returnValue };
    });
};
export const evaluateParsedConditions = (parsedConditions, data, value = null, required = null, def = false) => {
    for (const { conditions, returnValue } of parsedConditions) {
        const conditionResult = conditions.every((andConditions) => andConditions.some(({ left, operator, right }) => {
            const dataList = {
                _value: value,
                _required: required,
                ...data,
            };
            const leftValue = getNestedValue(dataList, left);
            return compareValues(operator, leftValue, right);
        }));
        if (conditionResult) {
            try {
                return JSON.parse(returnValue);
            }
            catch {
                return returnValue;
            }
        }
    }
    return def;
};
const getNestedValue = (data, path) => {
    return path.split('.').reduce((acc, part) => acc && acc[part], data);
};
export const compareValues = (operator, a, b) => {
    switch (operator) {
        case '=':
            return a === b;
        case '!=':
            return a !== b;
        case '>':
            return a > b;
        case '<':
            return a < b;
        case '>=':
            return a >= b;
        case '<=':
            return a <= b;
    }
};
export const isJson = (str) => {
    try {
        JSON.parse(str);
    }
    catch (e) {
        return false;
    }
    return true;
};
export const overwriteDefaults = (type, defaults) => {
    OVEWRITE_DEFAULT_OPTIONS[type] = defaults;
};
export const getOverwritenDefaults = (type, defaults) => {
    if (Object.keys(OVEWRITE_DEFAULT_OPTIONS).includes(type)) {
        return Object.assign({}, defaults, OVEWRITE_DEFAULT_OPTIONS[type]);
    }
    return defaults;
};
